{#
The Mia! Accounting Project
journal-entry-line-item-editor-modal: The modal of the journal entry line item editor

 Copyright (c) 2023 imacat.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

Author: imacat@mail.imacat.idv.tw (imacat)
First written: 2023/2/25
#}
<form id="accounting-line-item-editor">
  <input type="hidden" name="csrf_token" value="{{ csrf_token() }}">
  <div id="accounting-line-item-editor-modal" class="modal fade" tabindex="-1" aria-labelledby="accounting-line-item-editor-modal-label" aria-hidden="true">
    <div class="modal-dialog">
      <div class="modal-content">
        <div class="modal-header">
          <h1 class="modal-title fs-5" id="accounting-line-item-editor-modal-label">{{ A_("Line Item Content") }}</h1>
          <button type="button" class="btn-close" data-bs-dismiss="modal" aria-label="{{ A_("Close") }}"></button>
        </div>
        <div class="modal-body">
          <div id="accounting-line-item-editor-original-line-item-container" class="d-flex justify-content-between mb-3">
            <div class="accounting-line-item-editor-original-line-item-content">
              <div id="accounting-line-item-editor-original-line-item-control" class="form-control accounting-clickable accounting-material-text-field" data-bs-toggle="modal" data-bs-target="#accounting-original-line-item-selector-modal">
                <label class="form-label" for="accounting-line-item-editor-original-line-item">{{ A_("Original Line Item") }}</label>
                <div id="accounting-line-item-editor-original-line-item"></div>
              </div>
              <div id="accounting-line-item-editor-original-line-item-error" class="invalid-feedback"></div>
            </div>

            <div>
              <button id="accounting-line-item-editor-original-line-item-delete" class="btn btn-danger rounded-circle" type="button">
                <i class="fas fa-minus"></i>
              </button>
            </div>
          </div>

          <div class="mb-3">
            <div id="accounting-line-item-editor-description-control" class="form-control accounting-clickable accounting-material-text-field" data-bs-toggle="modal" data-bs-target="">
              <label class="form-label" for="accounting-line-item-editor-description">{{ A_("Description") }}</label>
              <div id="accounting-line-item-editor-description"></div>
            </div>
            <div id="accounting-line-item-editor-description-error" class="invalid-feedback"></div>
          </div>

          <div class="mb-3">
            <div id="accounting-line-item-editor-account-control" class="form-control accounting-clickable accounting-material-text-field" data-bs-toggle="modal" data-bs-target="">
              <label class="form-label" for="accounting-line-item-editor-account">{{ A_("Account") }}</label>
              <div id="accounting-line-item-editor-account"></div>
            </div>
            <div id="accounting-line-item-editor-account-error" class="invalid-feedback"></div>
          </div>

          <div class="form-floating mb-3">
            <input id="accounting-line-item-editor-amount" class="form-control" type="number" value="" min="0" max="" step="0.01" placeholder=" " required="required">
            <label for="accounting-line-item-editor-amount">{{ A_("Amount") }}</label>
            <div id="accounting-line-item-editor-amount-error" class="invalid-feedback"></div>
          </div>
        </div>
        <div class="modal-footer">
          <button type="button" class="btn btn-secondary" data-bs-dismiss="modal">{{ A_("Cancel") }}</button>
          <button type="submit" class="btn btn-primary">{{ A_("Save") }}</button>
        </div>
      </div>
    </div>
  </div>
</form>
