{#
The Mia! Accounting Project
ledger.html: The ledger

 Copyright (c) 2023 imacat.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

Author: imacat@mail.imacat.idv.tw (imacat)
First written: 2023/3/5
#}
{% extends "accounting/base.html" %}

{% block accounting_scripts %}
  <script src="{{ url_for("accounting.static", filename="js/material-fab-speed-dial.js") }}"></script>
  <script src="{{ url_for("accounting.static", filename="js/period-chooser.js") }}"></script>
{% endblock %}

{% block header %}{% block title %}{% if report.currency.code == accounting_default_currency_code() %}{{ A_("Ledger of %(account)s %(period)s", account=report.account.title, period=report.period.desc|title) }}{% else %}{{ A_("Ledger of %(account)s in %(currency)s %(period)s", currency=report.currency.name|title, account=report.account.title, period=report.period.desc|title) }}{% endif %}{% endblock %}{% endblock %}

{% block content %}

<div class="mb-3 accounting-toolbar">
  {% with use_currency_chooser = true,
          use_account_chooser = true,
          use_period_chooser = true %}
    {% include "accounting/report/include/toolbar-buttons.html" %}
  {% endwith %}
</div>

{% include "accounting/report/include/add-journal-entry-material-fab.html" %}

{% include "accounting/report/include/period-chooser.html" %}

{% include "accounting/report/include/search-modal.html" %}

{% if report.has_data %}
  {% with pagination = report.pagination %}
    {% include "accounting/include/pagination.html" %}
  {% endwith %}

  <div class="d-none d-md-block accounting-report-table {% if report.account.is_real %} accounting-ledger-real-table {% else %} accounting-ledger-nominal-table {% endif %}">
    <div class="accounting-report-table-header">
      <div class="accounting-report-table-row">
        <div>{{ A_("Date") }}</div>
        <div>{{ A_("Description") }}</div>
        <div class="accounting-amount">{{ A_("Debit") }}</div>
        <div class="accounting-amount">{{ A_("Credit") }}</div>
        {% if report.account.is_real %}
          <div class="accounting-amount">{{ A_("Balance") }}</div>
        {% endif %}
      </div>
    </div>
    <div class="accounting-report-table-body">
      {% if report.brought_forward %}
        {% with line_item = report.brought_forward %}
          <div class="accounting-report-table-row">
            {% include "accounting/report/include/ledger-row-desktop.html" %}
          </div>
        {% endwith %}
      {% endif %}
      {% for line_item in report.line_items %}
        <a class="accounting-report-table-row" href="{{ line_item.url|accounting_append_next }}">
          {% include "accounting/report/include/ledger-row-desktop.html" %}
        </a>
      {% endfor %}
    </div>
    {% if report.total %}
      {% with line_item = report.total %}
        <div class="accounting-report-table-footer">
          <div class="accounting-report-table-row">
            <div>{{ A_("Total") }}</div>
            <div class="accounting-amount">{{ line_item.debit|accounting_format_amount|accounting_default }}</div>
            <div class="accounting-amount">{{ line_item.credit|accounting_format_amount|accounting_default }}</div>
            {% if report.account.is_real %}
              <div class="accounting-amount {% if line_item.balance < 0 %} text-danger {% endif %}">{{ line_item.balance|accounting_report_format_amount }}</div>
            {% endif %}
          </div>
        </div>
      {% endwith %}
    {% endif %}
  </div>

  <div class="list-group d-md-none">
  {% if report.brought_forward %}
    {% with line_item = report.brought_forward %}
      <div class="list-group-item list-group-item-action d-flex justify-content-between">
        {% include "accounting/report/include/ledger-row-mobile.html" %}
      </div>
    {% endwith %}
  {% endif %}
  {% for line_item in report.line_items %}
    <a class="list-group-item list-group-item-action d-flex justify-content-between" href="{{ line_item.url|accounting_append_next }}">
      {% include "accounting/report/include/ledger-row-mobile.html" %}
    </a>
  {% endfor %}
  {% if report.total %}
    {% with line_item = report.total %}
      <div class="list-group-item list-group-item-action d-flex justify-content-between">
        {% include "accounting/report/include/ledger-row-mobile.html" %}
      </div>
    {% endwith %}
  {% endif %}
  </div>
{% else %}
  <p>{{ A_("There is no data.") }}</p>
{% endif %}

{% endblock %}
